#include "nvfs.h"

long nvfs_ioctl(struct file *filp, unsigned cmd, unsigned long arg)
{
	struct inode *inode = file_inode(filp);
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	unsigned flags, oldflags;
	int r;

	switch (cmd) {
		case FS_IOC_GETFLAGS:
			flags = nmi->i_flag & (FS_FL_USER_VISIBLE & NVFS_FLAGS);
			return put_user(flags, (int __user *)arg);
		case FS_IOC_SETFLAGS:
			r = mnt_want_write_file(filp);
			if (r)
				return r;

			if (!inode_owner_or_capable(inode)) {
				r = -EACCES;
				goto setflags_out;
			}

			if (get_user(flags, (int __user *) arg)) {
				r = -EFAULT;
				goto setflags_out;
			}

			flags = nvfs_mask_flags(inode->i_mode, flags);

			inode_lock(inode);
			if (IS_NOQUOTA(inode)) {
				r = -EPERM;
				goto setflags_unlock_out;
			}
			oldflags = nmi->i_flag;

			r = vfs_ioc_setflags_prepare(inode, oldflags, flags);
			if (r)
				goto setflags_unlock_out;

			flags = flags & (FS_FL_USER_MODIFIABLE & NVFS_FLAGS);
			flags |= oldflags & ~(FS_FL_USER_MODIFIABLE | NVFS_FLAGS);
			nmi->i_flag = flags;
			nv_store(nmi->nvi->i_flag, cpu_to_le32(flags));

			nvfs_set_inode_flags(inode);
			nvfs_update_time(inode, NULL, S_CTIME);
setflags_unlock_out:
			inode_unlock(inode);
setflags_out:
			mnt_drop_write_file(filp);
			return r;

		default:
			return -ENOTTY;
	}
}

#ifdef CONFIG_COMPAT
long nvfs_compat_ioctl(struct file *file, unsigned cmd, unsigned long arg)
{
	/* These are just misnamed, they actually get/put from/to user an int */
	switch (cmd) {
	case FS_IOC32_GETFLAGS:
		cmd = FS_IOC_GETFLAGS;
		break;
	case FS_IOC32_SETFLAGS:
		cmd = FS_IOC_SETFLAGS;
		break;
	default:
		return -ENOIOCTLCMD;
	}
	return nvfs_ioctl(file, cmd, (unsigned long)compat_ptr(arg));
}
#endif
